import os

from sqlalchemy import Column, BigInteger, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship

from wopmars.framework.database.Base import Base
from wopmars.framework.database.tables.IOPut import IOPut
from wopmars.utils.Logger import Logger


class IOFilePut(IOPut, Base):
    """
    This class extends IOPut and is specific to the input or output files. It is the model which store the references
    to the actual files needed by the user. The table ``wom_file`` associated with this model contains the
    following fields:

    - id: INTEGER - primary key - autoincrement - arbitrary ID
    - name: VARCHAR(255) - the name of the reference to the file
    - path: VARCHAR(255) - the path to the file
    - rule_id: INTEGER - foreign key to the associated rule ID: :class:`wopmars.framework.database.tables.ToolWrapper.ToolWrapper`
    - type_id: INTEGER - foreign key to the associated type ID: :class:`wopmars.framework.database.tables.Type.Type`
    - used_at: DATE - date at which the table have been used
    - size: INTEGER - the size of the file
    """
    __tablename__ = "wom_file"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))
    path = Column(String(255))
    rule_id = Column(Integer, ForeignKey("wom_rule.id"))
    type_id = Column(Integer, ForeignKey("wom_type.id"))
    used_at = Column(DateTime, nullable=True)
    size = Column(BigInteger, nullable=True)

    # One file is in Many rule_file and is in Many rule
    rule = relationship("ToolWrapper", back_populates="files", enable_typechecks=False)
    # One file has One type
    type = relationship("Type", back_populates="files")

    def is_ready(self):
        """
        Check if the file exists on the system.

        :return: boolean: True if it exists, false if not
        """
        Logger.instance().debug("Checking if " + self.name + " is ready: " + self.path)
        return os.path.isfile(self.path)

    def __eq__(self, other):
        return os.path.abspath(self.path) == os.path.abspath(other.path) and self.name == other.name

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return "<File (%s): %s: %s; size: %s; used_at: %s>" % (self.type.name, self.name, self.path, self.size, self.used_at)

    def __str__(self):
        return "file: " + self.name + ": " + self.path



import time, datetime
import random

from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class AddDateToPiece(ToolWrapper):
    __mapper_args__ = {
        "polymorphic_identity": "wopmars.example.AddDateToPiece"
    }

    def specify_input_table(self):
        return ["piece"]

    def specify_output_table(self):
        return ["piece"]

    def run(self):
        session = self.session()
        DatedPiece = self.output_table("piece")

        for p in self.session().query(DatedPiece).all():
            date = datetime.datetime.fromtimestamp(time.time() - random.randint(1000000, 100000000))
            p.date = date
            session.add(p)
        session.commit()


"""
Module containing the FooWrapper1 class
"""
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class FooWrapperNoRun(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapperNoRun"}

    def specify_input_file(self):
        return ["input1"]

    def specify_output_file(self):
        return ["output1"]

    def run(self):
        pass



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

# import models
import wopexamplesnp.model.SNP2Phenotype

import csv

class InsertSNP2Phenotype(ToolWrapper): # inherit WopMars
    __mapper_args__ = {'polymorphic_identity': __name__}
    #
    __input_file_snp2phenotype = "snp2phenotype"
    __input_table_snp = "SNP"
    __input_table_phenotype = "Phenotype"
    __output_table_snp2phenotype = "SNP2Phenotype"

    def specify_input_file(self):
        return [InsertSNP2Phenotype.__input_file_snp2phenotype]

    def specify_input_table(self):
        return [InsertSNP2Phenotype.__input_table_snp,
                InsertSNP2Phenotype.__input_table_phenotype]

    def specify_output_table(self):
        return [InsertSNP2Phenotype.__output_table_snp2phenotype]

    def run(self):
        session = self.session()
        engine = session._WopMarsSession__session.bind
        conn = engine.connect()
        #
        snp2phenotype_path = self.input_file(InsertSNP2Phenotype.__input_file_snp2phenotype)
        snp_model = self.input_table(InsertSNP2Phenotype.__input_table_snp)
        phenotype_model = self.input_table(InsertSNP2Phenotype.__input_table_phenotype)
        snp2phenotype_model = self.output_table(InsertSNP2Phenotype.__output_table_snp2phenotype)
        #
        # read input file
        input_file_obj_list = []
        for line in csv.reader(open(snp2phenotype_path, 'r', encoding='utf-8'), delimiter="\t"):
            snp_rsid = int(line[0])
            phenotype_name = line[1]
            input_file_snp_obj = {'rsid': snp_rsid}
            input_file_phenotype_obj = {'name': phenotype_name}
            try: # checks if exists Phenotype in db
                snp_id = session.query(snp_model).filter_by(**input_file_snp_obj).one().id
                phenotype_id = session.query(phenotype_model).filter_by(**input_file_phenotype_obj).one().id
                input_file_obj = {'snp_id': snp_id, 'phenotype_id': phenotype_id}
                try: # checks if exists Phenotype in db
                    session.query(snp2phenotype_model).filter_by(**input_file_obj).one()
                except: # if not, add
                    snp2phenotype_instance = snp2phenotype_model(**input_file_obj)
                    session.add(snp2phenotype_instance)
                session.commit()
            except: # if not, add
                print("warning: SNP rsid or phenotype name not in db")




import os
from setuptools import setup
from setuptools import find_packages

setup(
    name='wopexamplesnp',
    version='1.0',
    description='SNP example for experimenting Wopmars',
    author='Aitor Gonzalez - TAGC',
    author_email='aitor.gonzalez@univ-amu.fr',
    packages=find_packages(),
    )




from sqlalchemy.orm.query import Query

import wopmars.framework.database.SQLManager


class WopMarsQuery(Query):
    """
    This class extend the Query class from sqlalchemy.orm in purpose to use the lock provided by the SQLManager in order
    to query the database.

    It overide all methods relatives to finish a query:
      - all
      - one
      - first
      - count
      - one_or_none
      - scalar
    """
    # I call the whole thing in order to prevent circular import... thx python
    def all(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "all")

    def one(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "one")

    def first(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "first")

    def count(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "count")

    def one_or_none(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "one_or_none")

    def scalar(self):
        return wopmars.framework.database.SQLManager.SQLManager.instance().result_factory(self, "scalar")


"""
Module containing the OptionManager class.
"""
import os

from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.Singleton import SingletonMixin

from wopmars.constants import home_wopmars


class OptionManager(dict, SingletonMixin):
    """
    The OptionManager contains the command-line argument and can be retrieved from whereever in the software.

    OptionManager inherit from dict so it behaves exactly the same but takes only one possible argument.
    """

    def __init__(self, *args, **kwargs):
        """
        :param: dict_args: dictionnary containing the arguments passed to the command line.
        :return: void
        """
        super().__init__(*args, **kwargs)

    def validate(self, schema):
        self.validate_dir()
        self.validate_definition_file()
        self.validate_database()
        self.validate_dot()
        self.validate_log()
        self.validate_input_output_params()

        schema.validate(self)

        self.make_absolute_paths()

    def validate_input_output_params(self):
        if self["--input"] is None:
            self["--input"] = "{}"
        if self["--output"] is None:
            self["--output"] = "{}"
        if self["--params"] is None:
            self["--params"] = "{}"

    def validate_dir(self):
        if self["--directory"] == "$CWD":
            self["--directory"] = os.getcwd()

    def make_absolute_paths(self):
        if self["--wopfile"]:
            self["--wopfile"] = os.path.abspath(os.path.expanduser(self["--wopfile"]))
        if self["--log"]:
            self["--log"] = os.path.abspath(os.path.expanduser(self["--log"]))
        if self["--dot"]:
            self["--dot"] = os.path.abspath(os.path.expanduser(self["--dot"]))
        if self["--directory"]:
            self["--directory"] = os.path.abspath(os.path.expanduser(self["--directory"]))
        if self["--database"]:
            if self["--database"].split("://")[0]=="sqlite": # if sqlite replace with full path
                self["--database"]="sqlite:///" + os.path.abspath(os.path.expanduser(self["--database"].split(":///")[1]))
            # self["--database"] = {'db_connection': None, 'db_database': None, 'db_url': self["--database"]}
            # self["--database"]['db_connection'] = self["--database"]['db_url'].split("://")[0]
            # if self["--database"]['db_connection']=="sqlite":
            #     self["--database"]['db_database'] = os.path.abspath(os.path.expanduser(self["--database"]['db_url'].split(":///")[1]))

    def validate_definition_file(self):
        if self["--wopfile"] is None:
            self["--wopfile"] = "Wopfile"

    def validate_database(self):
        if self["--database"]:
            self["--database"] = os.path.expanduser(self["--database"])
        else:
            self["--database"] = os.path.join(self["--directory"], "sqlite:///db.sqlite")

    def validate_dot(self):
        if self["--dot"]:
            if self["--dot"][-1] == "/":
                self["--dot"] += "wopmars.dot"
            elif self["--dot"][-4:] != '.dot':
                self["--dot"] += ".dot"
            self["--dot"] = os.path.expanduser(self["--dot"])

    def validate_log(self):
        if self["--log"]:
            if self["--log"][0] == "$":
                self["--log"] = "~/.wopmars/wopmars.log"
            elif self["--log"][-1] == "/":
                self["--log"] += "wopmars.log"
            elif self["--log"][-4:] != '.log':
                self["--log"] += ".log"
            self["--log"] = os.path.expanduser(self["--log"])


    def __str__(self):
        s = ""

        s += "{\n\t"
        s += ",\n\t".join(["'" + str(k) + "': " + str(self[k]) for k in sorted(self.keys(), reverse=True)])
        s += "\n}"

        return s

    @staticmethod
    def initial_test_setup(mod_name="db"):
        if not os.path.isdir(home_wopmars):
            os.makedirs(home_wopmars)
        OptionManager.instance()["-v"] = 4
        OptionManager.instance()["--dot"] = None
        OptionManager.instance()["--log"] = os.path.join(os.path.expanduser("~"), ".wopmars/wopmars.log")
        OptionManager.instance()["--printtools"] = True
        OptionManager.instance()["--sourcerule"] = None
        OptionManager.instance()["--targetrule"] = None
        OptionManager.instance()["--forceall"] = None
        OptionManager.instance()["--dry-run"] = None
        OptionManager.instance()["tool"] = None
        OptionManager.instance()["--database"] = "sqlite:///" + os.path.join(PathFinder.get_module_path(), "test/output", mod_name + ".sqlite")
        OptionManager.instance()["--directory"] = PathFinder.get_module_path()
        OptionManager.instance()["--clear-history"] = False
        os.chdir(OptionManager.instance()["--directory"])
        OptionManager.instance()["--toolwrapper-log"] = False




from wopmars.framework.database.Base import Base
from sqlalchemy import Column, String, DateTime
from sqlalchemy.orm import relationship


class ModificationTable(Base):
    """
    The ModificationTable model contains the table names of the workflow and their date of last modification. The table
    ``wom_modification_table`` contains the following fields:

    - table_name: VARCHAR(255) - primary key - the name of the table
    - date: DATE - the date of last modification of the table
    """

    __tablename__ = "wom_modification_table"

    table_name = Column(String(255), primary_key=True)
    date = Column(DateTime, nullable=False)

    tables = relationship("IODbPut", back_populates="modification")

    def __repr__(self):
        return "<Modification on " + str(self.table_name) + ": " + str(self.date) + ">"



"""WopMars: Workflow Python Manager for Reproducible Science.

Usage:
  wopmars (-D DATABASE) (-w DEFINITION_FILE) [-n] [-p] [-F] [-v...] [-d DIR] [-g FILE] [-L FILE] [-f RULE | -t RULE] [-c] [-l]
  wopmars tool TOOLWRAPPER [-i DICT] [-o DICT] [-P DICT] [-p] [-F] [-D DATABASE] [-v...] [-d DIR] [-L FILE] [-g FILE] [-c] [-l]
  wopmars example [-d DIR]
  wopmars example_snp [-d DIR]

Arguments:
  DEFINITION_FILE  Path to the definition file of the workflow (Required)
  DATABASE         Path to the sqlite database file (Required)
  FILE             Path to a file.
  RULE             Name of a rule in the workflow definition file.
  TOOLWRAPPER      Path the the toolwrapper
  DICT             String formated like a dictionnary. Ex: "{'input1': 'path/to/input1', 'input2': 'path/to/input2'}"

Options:
  -D --database=DATABASE       REQUIRED: Set the path to the database, e.g -D sqlite:///db.sqlite
  -w --wopfile=DEFINITION_FILE REQUIRED: Set the path to the definition file.
  -c --clear-history           Clear WoPMaRS history. Should be used in case of bug which seem to be related to the history. Be carefull, clearing history will result in a re-execution of the whole workflow.
  -d --directory=DIR           Set the current working directory. Useful for working with relative paths [default: $CWD].
  -F --forceall                Force the execution of the workflow, without checking for previous executions.
  -f RULE --sourcerule=RULE    Execute the workflow from the given RULE.
  -g FILE --dot=FILE           Write dot representing the workflow in the FILE file (with .dot extension). This option needs to install WopMars with pygraphviz (pip install wopmars[pygraphviz])
  -h --help                    Show this help.
  -i --input=DICT              Set the input of the toolwrapper you want to use in the dictionnary format.
  -l --toolwrapper-log         Allow the toolwrapper to print its logs in the standard output.
  -L FILE --log=FILE           Write logs in FILE file [default: $HOME/.wopmars/wopmars.log].
  -n --dry-run                 Only display what would have been done.
  -o --output=DICT             Set the output of the toolwrapper you want to use in the dictionnary format.
  -P --params=DICT             Set the parameters of the toolwrapper you want to use in the dictionnary format.
  -p --printtools              Write logs in standard output.
  -t RULE --targetrule=RULE    Execute the workflow to the given RULE.
  -u --update                  Should be used when a file supposedly generated by the workflow already exists and should be used as it. (Not implemented)
  -v                           Set verbosity level, eg -v, -vv or -vvv

Example:
    wopmars example
    cd wopmars_example
    pip install .
    wopmars -w Wopfile -D "sqlite:///db.sqlite" -v -p -F
"""

import datetime
import os
import re
import sys
import time

from docopt import docopt, DocoptExit
from schema import Schema, And, Or, Use, SchemaError

from wopmars.example.ExampleBuilder import ExampleBuilder
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.management.WorkflowManager import WorkflowManager
from wopmars.utils.DictUtils import DictUtils
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.exceptions.WopMarsException import WopMarsException
from wopmars.constants import home_wopmars

# todo combinatorices pour les rules
# todo option pour reset les resultats (supprimer le contenu de la database) / fresh run
# todo ajouter un flag NOT_FINISHED aux executions

__version__='0.0.2'

class WopMars:

    @staticmethod
    def run(argv):
        """
        Entry-point of the program
        """

        # if the command line is malformed, docopt interrupt the software.
        try:
            if argv[1:] == []: # If not arguments, run the help
                argv.append('-h')
            OptionManager.instance().update(docopt(__doc__, argv=argv[1:]))
        except DocoptExit as SE:
            print("Bad argument in the command line: \n\t" + " ".join(argv) + "\n" + str(SE))
            sys.exit(2)
        try:
            schema_option = Schema({
                '--wopfile': Or("Wopfile", str),
                '--database': Use(PathFinder.check_database_valid_url),
                '-v': Or(0, And(int, lambda n: 1 <= n <= 2)),
                '--dot': Or(None, And(Use(PathFinder.check_valid_path), Use(PathFinder.check_pygraphviz))),
                "--log": Use(PathFinder.check_valid_path),
                '--printtools': Use(bool),
                "--sourcerule": Or(None, str),
                "--targetrule": Or(None, str),
                "--forceall": Use(bool),
                "--dry-run": Use(bool),
                "--directory": Use(PathFinder.create_workingdir),
                "--input": Use(DictUtils.str_to_dict),
                "--output": Use(DictUtils.str_to_dict),
                "--params": Use(DictUtils.str_to_dict),
                "TOOLWRAPPER": Or(None, Use(PathFinder.is_in_python_path)),
                "tool": Use(bool),
                "example": Use(bool),
                "example_snp": Use(bool),
                "--clear-history": Use(bool),
                "--toolwrapper-log": Use(bool)
            })
            # The option values are validated using schema library
            OptionManager.instance().validate(schema_option)
            os.chdir(OptionManager.instance()["--directory"])
        except SchemaError as schema_msg:
            Logger.instance().debug("\nCommand line Args:" + str(OptionManager.instance()))
            # regex for the different possible error messages.
            match_open_def = re.match(r"^open\('(.[^\)]+)'\)", str(schema_msg))
            match_dot_def = re.match(r"^check_valid_path\(('.[^\)]+')\)", str(schema_msg))
            match_wrong_key = re.match(r"^Wrong keys ('.[^\)]+')", str(schema_msg))
            match_pygraphviz = re.match(r".*dot.*", str(schema_msg))
            print(match_pygraphviz)
            # Check the different regex..
            if match_open_def:
                Logger.instance().error("The file " + match_open_def.group(1) + " cannot be opened. It may not exist.")
            elif match_dot_def:
                Logger.instance().error("The path " + match_dot_def.group(1) + " is not valid.")
            elif match_wrong_key:
                # Normally never reach
                Logger.instance().error("The option key " + match_wrong_key.group(1) + " is not known.")
            elif match_pygraphviz:
                Logger.instance().error("The dot file path is not valid or the pygraphviz module is not installed. In the second case, install wopmars with pygraphviz: pip install wopmars[pygraphviz]")
            else:
                # Normally never reach
                Logger.instance().error("An unknown error has occured. Message: " + str(schema_msg))
            sys.exit(2)

        Logger.instance().debug("\nCommand line Args:" + str(OptionManager.instance()))

        if OptionManager.instance()["example"]:
            ExampleBuilder().build()
            sys.exit(1)

        if OptionManager.instance()["example_snp"]:
            ExampleBuilder().build_snp()
            sys.exit(1)


        wm = WorkflowManager()
        try:
            wm.run()
        except WopMarsException as WE:
            Logger.instance().error(str(WE))
            session = SQLManager.instance().get_session()
            try:
                finished_at = datetime.datetime.fromtimestamp(time.time())
                Logger.instance().error("The workflow has encountered an error at: " + str(finished_at))
                wm.set_finishing_informations(finished_at, "ERROR")
            except AttributeError:
                session.rollback()
                Logger.instance().error("The execution has not even begun. No informations will be stored in the database.")
            except Exception as e:
                Logger.instance().error("An error occured during the rollback of the changement of the database which can be now unstable:" +
                                        str(e))
            sys.exit(1)
        # except Exception as e:
        #     Logger.instance().error("An unknown error has occured:\n" + str(e))
        #     sys.exit(1)


def run():
    # todo ask lionel: du coup, ça, c'est pas censé y être? vu que les toolwrappers ne devront pas être dans
    # l'arborescence du projet? même les toolwrappers de base
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/toolwrappers/")
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + "/base/")



    if not os.path.isdir(home_wopmars):
        os.makedirs(home_wopmars)

    WopMars().run(sys.argv)

if __name__ == "__main__":
    run()




"""Simple reader-writer locks in Python
Many readers can hold the lock XOR one and only one writer"""
import threading

version = """$Id: 04-1.html,v 1.3 2006/12/05 17:45:12 majid Exp $"""


class RWLock:
    """
    A simple reader-writer lock Several readers can hold the lock
    simultaneously, XOR one writer. Write locks have priority over reads to
    prevent write starvation.
    """
    def __init__(self):
        self.lockrw = 0
        self.writers_waiting = 0
        self.monitor = threading.Lock()
        self.readers_ok = threading.Condition(self.monitor)
        self.writers_ok = threading.Condition(self.monitor)

    def acquire_read(self):
        """Acquire a read lock. Several threads can hold this typeof lock.
        It is exclusive with write locks."""
        self.monitor.acquire()
        while self.lockrw < 0 or self.writers_waiting:
            self.readers_ok.wait()
        self.lockrw += 1
        self.monitor.release()

    def acquire_write(self):
        """Acquire a write lock. Only one thread can hold this lock, and
        only when no read locks are also held."""
        self.monitor.acquire()
        while self.lockrw != 0:
            self.writers_waiting += 1
            self.writers_ok.wait()
            self.writers_waiting -= 1
        self.lockrw = -1
        self.monitor.release()

    def promote(self):
        """Promote an already-acquired read lock to a write lock
        WARNING: it is very easy to deadlock with this method"""
        self.monitor.acquire()
        self.lockrw -= 1
        while self.lockrw != 0:
            self.writers_waiting += 1
            self.writers_ok.wait()
            self.writers_waiting -= 1
        self.lockrw = -1
        self.monitor.release()

    def demote(self):
        """Demote an already-acquired write lock to a read lock"""
        self.monitor.acquire()
        self.lockrw = 1
        self.readers_ok.notifyAll()
        self.monitor.release()

    def release(self):
        """Release a lock, whether read or write."""
        self.monitor.acquire()
        if self.lockrw < 0:
            self.lockrw = 0
        else:
            self.lockrw -= 1
        wake_writers = self.writers_waiting and self.lockrw == 0
        wake_readers = self.writers_waiting == 0
        self.monitor.release()
        if wake_writers:
            self.writers_ok.acquire()
            self.writers_ok.notify()
            self.writers_ok.release()
        elif wake_readers:
            self.readers_ok.acquire()
            self.readers_ok.notifyAll()
            self.readers_ok.release()

if __name__ == '__main__':
    import time
    rwl = RWLock()
    class Reader(threading.Thread):
        def run(self):
            print( self, 'start')
            rwl.acquire_read()
            print( self, 'acquired')
            time.sleep(5)
            print( self, 'stop')
            rwl.release()
    class Writer(threading.Thread):
        def run(self):
            print( self, 'start')
            rwl.acquire_write()
            print( self, 'acquired')
            time.sleep(10)
            print( self, 'stop')
            rwl.release()
    class ReaderWriter(threading.Thread):
        def run(self):
            print( self, 'start')
            rwl.acquire_read()
            print( self, 'acquired')
            time.sleep(5)
            rwl.promote()
            print( self, 'promoted')
            time.sleep(5)
            print( self, 'stop')
            rwl.release()
    class WriterReader(threading.Thread):
        def run(self):
            print( self, 'start')
            rwl.acquire_write()
            print( self, 'acquired')
            time.sleep(10)
            print( self, 'demoted')
            rwl.demote()
            time.sleep(10)
            print( self, 'stop')
            rwl.release()
    Reader().start()
    time.sleep(1)
    Reader().start()
    time.sleep(1)
    ReaderWriter().start()
    time.sleep(1)
    WriterReader().start()
    time.sleep(1)
    Reader().start()


from sqlalchemy import Column, Integer, DateTime, Float, String
from sqlalchemy.orm import relationship

from wopmars.framework.database.Base import Base


class Execution(Base):
    """
    The execution table is a counter of the executions. It allows to discriminate one execution from one other.

    If 2 executions are exactly the same, they will be both stocked in the database with different execution ids.
    The table ``wom_execution`` contains the following fields:

    - id: INTEGER: primary key - auto increment - arbitrary iD
    - started_at: DATE - date at wich the execution started
    - finished_at: DATE - date at wich the execution finished
    - time: FLOAT - the total time of execution
    - status: VARCHAR(255) - the final status of the Execution. it can be:

        - SUCCESS
        - ERROR

    """

    __tablename__ = "wom_execution"

    id = Column(Integer, primary_key=True, autoincrement=True)
    started_at = Column(DateTime, nullable=True)
    finished_at = Column(DateTime, nullable=True)
    time = Column(Float, nullable=True)
    status = Column(String(255), nullable=True)

    # One execution has many rules
    rules = relationship("ToolWrapper", back_populates="execution")




"""
Module containing the ToolThread class.
"""
import datetime
import errno
import threading
import os
import time
import traceback

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.management.Observable import Observable
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class ToolThread(threading.Thread, Observable):
    """
    The class ToolThread is a wrapper for executing toolwrappers.

    It has been designed in order to implement the multithreading, this is why it inherit from threading.Thread.
    """

    def __init__(self, toolwrapper):
        """
        self.__dry = True means that the tool shouldn't be executed because its output already exist

        :return:
        """
        threading.Thread.__init__(self)
        self.__set_observer = set([])
        # the wrapped toolwrapper
        self.__toolwrapper = toolwrapper
        #self.__dry is different than the --dry-run option because it says "this has already been executed" whereas
        # the --dry-run option means "simulate the whole execution"
        # self.__dry can be True even if the --dry-run mode is enabled: it means "this tool has already its output, you
        # don't even need to simulate execution, just skip"
        self.__dry = False

    def get_toolwrapper(self):
        return self.__toolwrapper

    def set_dry(self, dry):
        self.__dry = dry

    def get_dry(self):
        return self.__dry

    def run(self):
        """
        Run the tool and fire events.
        :return:
        """

        session_tw = SQLManager.instance().get_session()
        start = datetime.datetime.fromtimestamp(time.time())
        try:
            self.__toolwrapper.set_session(session_tw)
            # if the tool need to be executed because its output doesn't exist
            if not self.__dry:
                Logger.instance().info(
                    "\n" + str(self.__toolwrapper) + "\n" + "command line: \n\t" + self.get_command_line())
                # if you shouldn't simulate
                if not OptionManager.instance()["--dry-run"]:
                    Logger.instance().info("Rule: " + str(self.__toolwrapper.name) + " -> " + self.__toolwrapper.__class__.__name__ + " started.")
                    # mkdir -p output dir: before running we need output dir
                    output_file_fields = self._ToolThread__toolwrapper.specify_output_file()
                    for out_field in output_file_fields:
                        out_file_path = self._ToolThread__toolwrapper.output_file(out_field)
                        out_dir = os.path.dirname(out_file_path)
                        try:
                            os.makedirs(out_dir)
                        except OSError as exception:
                            if exception.errno != errno.EEXIST:
                                raise
                    # end of mkdir -p output dir
                    self.__toolwrapper.run()
                    session_tw.commit()
                    self.__toolwrapper.set_execution_infos(start, datetime.datetime.fromtimestamp(time.time()), "EXECUTED")
                else:
                    Logger.instance().debug("Dry-run mode enabled. Execution skiped.")
                    self.__toolwrapper.set_execution_infos(status="DRY")
            else:
                Logger.instance().info("Rule: " + str(self.__toolwrapper.name) + " -> " + self.__toolwrapper.__class__.__name__ + " skiped.")
                self.__toolwrapper.set_execution_infos(start, datetime.datetime.fromtimestamp(time.time()), "ALREADY_EXECUTED")
        except Exception as e:
            session_tw.rollback()
            self.__toolwrapper.set_execution_infos(start, datetime.datetime.fromtimestamp(time.time()), "EXECUTION_ERROR")
            raise WopMarsException("Error while executing rule " + self.__toolwrapper.name +
                                   " (ToolWrapper " + self.__toolwrapper.toolwrapper + ")",
                                   "Full stack trace: \n" + str(traceback.format_exc()))
        finally:
            # todo twthread , fermer session
            # session_tw.close()
            pass
        self.fire_success()

    def get_command_line(self):
        """
        This create a string containing the command line for executing the toolwrapper only.

        :return: The string containg the command line
        """
        list_str_inputs_files = [f.name + "': '" + f.path for f in self.__toolwrapper.files if f.type.name == "input"]
        list_str_inputs_tables = [t.tablename + "': '" + t.model for t in self.__toolwrapper.tables if t.type.name == "input"]
        str_input_dict = ""
        str_input_dict_files = ""
        str_input_dict_tables = ""

        if list_str_inputs_files:
            str_input_dict_files = "'file':{'" + "', '".join(list_str_inputs_files) + "'}"
        if list_str_inputs_tables:
            str_input_dict_tables = "'table':{'" + "', '".join(list_str_inputs_tables) + "'}"
        if list_str_inputs_files or list_str_inputs_tables:
            str_input_dict = " -i \"{%s}\"" % (", ".join([s for s in [str_input_dict_files, str_input_dict_tables] if s != ""]))

        list_str_outputs_files = [f.name + "': '" + f.path for f in self.__toolwrapper.files if f.type.name == "output"]
        list_str_outputs_tables = [t.tablename + "': '" + t.model for t in self.__toolwrapper.tables if t.type.name == "output"]
        str_output_dict = ""
        str_output_dict_files = ""
        str_output_dict_tables = ""

        if list_str_outputs_files:
            str_output_dict_files = "'file':{'" + "', '".join(list_str_outputs_files) + "'}"
        if list_str_outputs_tables:
            str_output_dict_tables = "'table':{'" + "', '".join(list_str_outputs_tables) + "'}"
        if list_str_outputs_files or list_str_outputs_tables:
            str_output_dict = " -o \"{%s}\"" % (", ".join([s for s in [str_output_dict_files, str_output_dict_tables] if s != ""]))

        list_str_params = []
        str_params_dict = ""

        if list_str_params:
            str_params_dict = " -P \"{'" + "', '".join(list_str_params) + "'}\""

        consistent_keys = ["--forceall", "--dot", "--log", ]
        s = ""
        s += "wopmars tool " + self.__toolwrapper.toolwrapper + str_input_dict + str_output_dict + str_params_dict + " " + \
             " ".join(str(key) + " " + str(OptionManager.instance()[key]) for key in OptionManager.instance().keys() if key in consistent_keys and OptionManager.instance()[key] is not None and type(OptionManager.instance()[key]) != bool) + \
             " " + " ".join(str(key) for key in OptionManager.instance().keys() if key in consistent_keys and OptionManager.instance()[key] is True and type(OptionManager.instance()[key]) == bool)

        return s

    def get_observers(self):
        """
        Return the set of observers.

        :return: set observers
        """
        return self.__set_observer

    def subscribe(self, obs):
        """
        An observer subscribes to the obervable.

        :param obs:
        :return:
        """
        self.__set_observer.add(obs)

    def fire_failure(self):
        """
        Notify all ToolWrapperObservers that the execution has failed.

        :return:
        """
        for obs in self.get_observers():
            obs.notify_failure(self)

    def fire_success(self):
        """
        Notify all ToolWrapperObservers that the run has suceeded

        :return:
        """
        for obs in self.get_observers():
            obs.notify_success(self)

    def __eq__(self, other):
        assert isinstance(other, self.__class__)
        return self.__toolwrapper == other.get_toolwrapper()

    def __hash__(self):
        return id(self)



import os
import unittest
from unittest import TestCase

from test.resource.model.FooBase import FooBase
from test.resource.model.FooBase2 import FooBase2
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder


class TestIODbPut(TestCase):
    def setUp(self):
        self.s_root_path = PathFinder.get_module_path()
        OptionManager.initial_test_setup()
        print(OptionManager.instance()["--log"])

        SQLManager.instance().create_all()
        self.__local_session = SQLManager.instance().get_session()
        try:
            for i in range(10):
                self.__local_session.add(FooBase(name="testIODB " + str(i)))
            self.__local_session.commit()
        except Exception as e:
            self.__local_session.rollback()
            self.__local_session.close()
            raise e

        self.__io_base_existing = IODbPut(model="FooBase", tablename="FooBase")
        self.__io_base_existing.set_table(FooBase)
        self.__io_base_existing2 = IODbPut(model="FooBase", tablename="FooBase")
        self.__io_base_existing2.set_table(FooBase)
        self.__io_base_existing3 = IODbPut(model="FooBase2", tablename="FooBase2")
        self.__io_base_existing3.set_table(FooBase2)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove(os.path.join(self.s_root_path, "test/output"))
        OptionManager._drop()
        SQLManager._drop()

    def test_eq(self):
        self.assertEqual(self.__io_base_existing, self.__io_base_existing2)
        self.assertNotEqual(self.__io_base_existing, self.__io_base_existing3)

    def test_is_ready(self):
        self.assertTrue(self.__io_base_existing.is_ready())
        self.assertFalse(self.__io_base_existing3.is_ready())

if __name__ == "__main__":
    unittest.main()



"""
Module containing the FooWrapper1 class
"""
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

from sqlalchemy.sql import select


class FooWrapperCore(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {'polymorphic_identity': "FooWrapperCore"}

    def specify_output_table(self):
        return ["FooBase7"]

    def run(self):
        foobase = self.output_table("FooBase7").__table__
        inserted_list = []
        for i in range(10000):
            inserted_list.append({'id': i, 'name': "FooWrapperCore " + str(i)})
        ins = foobase.insert()
        self.session().execute(ins, inserted_list)

        s = select([foobase])
        result = self.session().execute(s)
        for row in result:
            print(row)



"""
Module containing the ToolWrapperObserver class
"""


class ToolWrapperObserver:
    """
    Observers of ToolWrappers should implement this interface.
    """

    def notify_success(self, toolwrapper):
        """
        Behavior of the Observer when the execution of toolwrapper has succeeded.
        :return:
        """
        raise NotImplementedError

    def notify_failure(self, toolwrapper):
        """
        Behavior of the Observer when the execution of toolwrapper has failed.
        :return:
        """
        raise NotImplementedError



"""
Module containing the Observable class
"""


class Observable:
    """
    class Observable
    """    
    def get_observers(self):
        """

        :return: set observers
        """
        raise NotImplementedError

    def subscribe(self, obs):
        raise NotImplementedError

    def fire_failure(self):
        raise NotImplementedError

    def fire_success(self):
        raise NotImplementedError




from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

# import models
import wopexamplesnp.model.Phenotype

class InsertPhenotype(ToolWrapper): # inherit WopMars
    __mapper_args__ = {'polymorphic_identity': __name__}
    __input_file_snp2phenotype = "snp2phenotype"
    __output_table_phenotype = "Phenotype"

    def specify_input_file(self):
        return [InsertPhenotype.__input_file_snp2phenotype]

    def specify_output_table(self):
        return [InsertPhenotype.__output_table_phenotype]

    def run(self):
        session = self.session()
        snp2phenotype_path = self.input_file(InsertPhenotype.__input_file_snp2phenotype)
        phenotype_model = self.output_table(InsertPhenotype.__output_table_phenotype)
        with open(snp2phenotype_path, "r") as fin:
            for snp2phenotype_line in fin.readlines():
                phenotype_name = snp2phenotype_line.strip().split("\t")[1]
                input_file_obj = {'name': phenotype_name}
                try: # checks if exists Phenotype in db
                    session.query(phenotype_model).filter_by(**input_file_obj).one()
                except: # if not, add
                    session.add(phenotype_model(**input_file_obj))
                session.commit()





"""
Module containing the WopMarsSession class.
"""
from sqlalchemy.sql.elements import ClauseElement

from wopmars.framework.database.WopMarsQuery import WopMarsQuery
from wopmars.utils.Logger import Logger


class WopMarsSession:
    """
    The class WopMarsSession is used in order to allow the Toolwrapper Developer to access the database.

    This class is a pivot between the Toolwrapper Developer and the SQLManager. Every call to the database has to pass
    through the SQLManager but the developer has only access to the session.
    """

    def __init__(self, session, manager):
        """
        :param session: contains a SQLAlchemy session
        :param manager: contains the SQLManager
        """
        self.__manager = manager
        self.__session = session

    def commit(self):
        """
        Validate changes on the database. Should be used when everything is ok.
        """
        if self.something():
            Logger.instance().debug(str(self.__session) + " is about to commit.")
            Logger.instance().debug("Operations to be commited in session" + str(self.__session) + ": \n\tUpdates:\n\t\t" +
                                    "\n\t\t".join([str(k) for k in self.__session.dirty]) +
                                    "\n\tInserts:\n\t\t" +
                                    "\n\t\t".join([str(k) for k in self.__session.new]))
            # call on SQLManager commit method to use the lock
            self.__manager.commit(self.__session)

    def rollback(self):
        """
        Rollback changes on the database. Should be used in case of error.
        :return:
        """
        Logger.instance().debug("Operations to be rollbacked in session" + str(self.__session) + ": \n\tUpdates:\n\t\t" +
                                "\n\t\t".join([str(k) for k in self.__session.dirty]) +
                                "\n\tInserts:\n\t\t" +
                                "\n\t\t".join([str(k) for k in self.__session.new]))
        # call on SQLManager commit method to use the lock
        self.__manager.rollback(self.__session)

    def query(self, table):
        """
        Return a WopMarsQuery used for querying database using the ORM.

        :param table: The mapper object on which you want to query.
        :return: WopMarsQuery on the desired table and bind to the current session.
        """
        return WopMarsQuery(table, session=self.__session)

    def add(self, item):
        """
        Add an item to the current session. To write the change on the database, don't forget to commit.

        :param item: a mapper object
        """
        self.__session.add(item)

    def add_all(self, collection):
        """
        Add a collection of items to the current session. To write the change on the database, don't forget to commit.

        :param collection: collection of mappers
        """
        self.__session.add_all(collection)

    def delete(self, entry):
        """
        Delete an entry from the database. To write the change on the database, don't forget to commit.

        :param entry: A mapper object already existent in the database.
        """
        self.__session.delete(entry)

    def delete_content(self, table):
        """
        Delete the content of a given table.

        :param table: A mapper object representing the table in which the content should be deleted.
        """
        Logger.instance().debug("Deleting content of table " + table.__tablename__ + "...")
        self.__manager.execute(self.__session, table.__table__.delete())
        Logger.instance().debug("Content of table " + table.__tablename__ + " deleted.")

    def execute(self, statement, *args, **kwargs):
        """
        Execute a statement of the given values.

        :param statement: SQLAlchemy statement (insert, delete, etc..)
        :param args: dict of values
        :param kwargs: exploded dict of values
        :return: return the result of the execution of the statement
        """
        Logger.instance().debug("WopMarsSession.execute(" + str(statement) + ", " + str(args) + ", " + str(kwargs) + ")")
        return self.__manager.execute(self.__session, statement, *args, **kwargs)

    def close(self):
        """
        Close the current session. Shouldn't be used.

        :return:
        """
        self.__session.close()

    def _session(self):
        """
        Return the SQLAlchemy session shouldn't be used by the ToolWrapper Developper.
        """
        return self.__session

    def something(self):
        """
        Check if there is something in the session.

        :return: Bool saying if there is something in the session.
        """
        return bool(self.__session.new) or bool(self.__session.dirty)

    def get_or_create(self, model, defaults=None, **kwargs):
        """
        This method allows to check if an entry exist in database and if not, create it. It also return a boolean
        which says if the entry has been created (True) or not (False)

        :param model: A SQLAlchemy model
        :param defaults: Dict containing the default parameter to use in fields if the entry doesn't exist.
        :param kwargs: Specify the values of the fields which are looked for. Create them if not.

        :return: tuple: (instance of the created entry, boolean)
        """
        instance = self.__session.query(model).filter_by(**kwargs).first()
        if instance:
            return instance, False
        else:
            params = dict((k, v) for k, v in kwargs.items() if not isinstance(v, ClauseElement))
            params.update(defaults or {})
            instance = model(**params)
            self.__session.add(instance)
            return instance, True

    # def pandas_to_sql(self, df, tablename, *args, **kwargs):
    #     """
    #     Interface for using the to_sql method from pandas dataframes.
    #
    #     :param df: pandas dataframe
    #     :type df: DataFrame
    #     :param tablename: the name of the table on which you want to perform operations
    #     :type tablename: str
    #     :param args: args of the conventionnal pandas.to_sql method
    #     :param kwargs: kwargs of the conventionnal pandas.to_sql method
    #     """
    #     self.__manager.pandas_to_sql(df, tablename, *args, **kwargs)
    #
    # def pandas_read_sql(self, sql, *args, **kwargs):
    #     """
    #     Interface for using the read_sql method from pandas dataframes.
    #
    #     :param sql: the sql query to perform on the database
    #     :type sql: str
    #     :param args: args of the conventionnal pandas.read_sql method
    #     :param kwargs: kwargs of the conventionnal pandas.read_sql method
    #     :return: Pandas dataframe contianing the result
    #     """
    #     return self.__manager.pandas_read_sql(sql, *args, **kwargs)




"""
Module containing the FooWrapper1 class
"""
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class FooWrapperH(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    __mapper_args__ = {
        "polymorphic_identity": "FooWrapperH"
    }

    def specify_input_table(self):
        return ["FooBaseH"]

    def specify_output_table(self):
        return ["FooBaseH"]

    def run(self):
        q = self.session().query(self.input_table("FooBaseH")).all()
        for r in q:
            r.name2="FooBaseH2 - " + str(r.id)
            self.session().add(r)
        self.session().commit()



import datetime
import os
import subprocess
import time
import unittest
from unittest import TestCase

from test.resource.model.FooBase import FooBase
from test.resource.wrapper.FooWrapper2 import FooWrapper2
from test.resource.wrapper.FooWrapper3 import FooWrapper3
from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IODbPut import IODbPut
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.framework.database.tables.ModificationTable import ModificationTable
from wopmars.framework.database.tables.Option import Option
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper
from wopmars.framework.database.tables.Type import Type
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class TestToolWrapper(TestCase):
    def setUp(self):
        OptionManager.initial_test_setup()
        self.s_root_path = PathFinder.get_module_path()
        SQLManager.instance().create_all()

        set_tw_to_add = set()
        self.__session = SQLManager.instance().get_session()

        self.input_entry = Type(name="input")
        self.output_entry = Type(name="output")

        ### Toolwrappers for __eq__ test_bak
        opt1 = Option(name="param1", value="1")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper1 = ToolWrapper(rule_name="rule1")
        self.__toolwrapper1.files.extend([f1, f2])
        self.__toolwrapper1.options.append(opt1)

        opt1 = Option(name="param1", value="1")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper2 = ToolWrapper(rule_name="rule2")
        self.__toolwrapper2.files.extend([f1, f2])
        self.__toolwrapper2.options.append(opt1)

        opt1 = Option(name="param2", value="2")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper3 = ToolWrapper(rule_name="rule3")
        self.__toolwrapper3.files.extend([f1, f2])
        self.__toolwrapper3.options.append(opt1)

        ### ToolWrappers for content_respected
        opt1 = Option(name="param1", value="2")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1
        t1.type = self.input_entry

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2
        t2.type = self.output_entry

        self.__foowrapper_right_content = FooWrapper3(rule_name="rule1")
        self.__foowrapper_right_content.files.extend([f1, f2])
        self.__foowrapper_right_content.tables.extend([t1, t2])
        self.__foowrapper_right_content.options.append(opt1)

        opt1 = Option(name="param1", value="String")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2

        self.__foowrapper_wrong_content1 = FooWrapper3(rule_name="rule2")
        self.__foowrapper_wrong_content1.files.extend([f1, f2])
        self.__foowrapper_wrong_content1.tables.extend([t1, t2])
        self.__foowrapper_wrong_content1.options.append(opt1)

        opt1 = Option(name="param2", value="2")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        f3 = IOFilePut(name="input2", path="file2.txt")
        f3.type = self.input_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2

        self.__foowrapper_wrong_content2 = FooWrapper3(rule_name="rule3")
        self.__foowrapper_wrong_content2.files.extend([f1, f2, f3])
        self.__foowrapper_wrong_content2.tables.extend([t1, t2])
        self.__foowrapper_wrong_content2.options.append(opt1)

        opt1 = Option(name="param2", value="2")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2

        self.__foowrapper_wrong_content3 = FooWrapper3(rule_name="rule3")
        self.__foowrapper_wrong_content3.files.extend([f1, f2])
        self.__foowrapper_wrong_content3.tables.extend([t1, t2])
        self.__foowrapper_wrong_content3.options.append(opt1)

        opt1 = Option(name="param1", value="String")

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2

        self.__foowrapper_wrong_content4 = FooWrapper3(rule_name="rule3")
        self.__foowrapper_wrong_content4.files.extend([f1, f2])
        self.__foowrapper_wrong_content4.tables.extend([t1, t2])
        self.__foowrapper_wrong_content4.options.append(opt1)

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.table = t1

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.table = t2

        self.__foowrapper_wrong_content5 = FooWrapper3(rule_name="rule3")
        self.__foowrapper_wrong_content5.files.extend([f1, f2])
        self.__foowrapper_wrong_content5.tables.extend([t1, t2])

        ### TooLWrappers for follows

        f1 = IOFilePut(name="input1", path="file1.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper_first = FooWrapper2(rule_name="rule1")
        self.__toolwrapper_first.files.extend([f1, f2])

        f1 = IOFilePut(name="input1", path="file2.txt")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file3.txt")
        f2.type = self.output_entry

        self.__toolwrapper_second = FooWrapper2(rule_name="rule2")
        self.__toolwrapper_second.files.extend([f1, f2])

        ### ToolWrappers for are_input_ready

        s_root_path = PathFinder.get_module_path()

        s_path_to_example_file_that_exists = os.path.join(s_root_path, "test/resource/input_files/input_file1.txt")

        f1 = IOFilePut(name="input1", path=s_path_to_example_file_that_exists)
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper_ready = FooWrapper2(rule_name="rule2")
        self.__toolwrapper_ready.files.extend([f1, f2])

        f1 = IOFilePut(name="input1", path="/not/existent/file")
        f1.type = self.input_entry

        f2 = IOFilePut(name="output1", path="file2.txt")
        f2.type = self.output_entry

        self.__toolwrapper_not_ready = FooWrapper2(rule_name="rule2")
        self.__toolwrapper_not_ready.files.extend([f1, f2])

    def test_eq(self):
        self.assertEqual(self.__toolwrapper1, self.__toolwrapper2)
        self.assertNotEqual(self.__toolwrapper1, self.__toolwrapper3)

    def test_is_content_respected(self):
        try:
            self.__foowrapper_right_content.is_content_respected()
        except WopMarsException:
            raise AssertionError('Should not raise exception')

        self.assertRaises(WopMarsException, self.__foowrapper_wrong_content1.is_content_respected)
        self.assertRaises(WopMarsException, self.__foowrapper_wrong_content2.is_content_respected)
        self.assertRaises(WopMarsException, self.__foowrapper_wrong_content3.is_content_respected)
        self.assertRaises(WopMarsException, self.__foowrapper_wrong_content4.is_content_respected)
        self.assertRaises(WopMarsException, self.__foowrapper_wrong_content5.is_content_respected)

    def test_follows(self):
        self.assertTrue(self.__toolwrapper_second.follows(self.__toolwrapper_first))
        self.assertFalse(self.__toolwrapper_first.follows(self.__toolwrapper_second))

    def test_are_inputs_ready(self):
        self.assertTrue(self.__toolwrapper_ready.are_inputs_ready())
        self.assertFalse(self.__toolwrapper_not_ready.are_inputs_ready())

        SQLManager.instance().get_session().add_all([FooBase(name="test_bak " + str(i)) for i in range(5)])
        SQLManager.instance().get_session().commit()

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)

        t1.type = self.input_entry

        toolwrapper_ready2 = FooWrapper2(rule_name="rule2")
        toolwrapper_ready2.tables.append(t1)
        self.assertTrue(toolwrapper_ready2.are_inputs_ready())

        SQLManager.instance().drop(FooBase.__tablename__)

        self.assertFalse(toolwrapper_ready2.are_inputs_ready())

    def test_same_input_than(self):

        moment = datetime.datetime.fromtimestamp(time.time())

        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.type = self.input_entry
        modif = ModificationTable(table_name="FooBase", date=moment)
        modif.tables.append(t1)

        t2 = IODbPut(model="FooBase", tablename="FooBase")
        t2.set_table(FooBase)
        t2.type = self.input_entry
        modif = ModificationTable(table_name="FooBase", date=moment)
        modif.tables.append(t2)

        f1 = IOFilePut(name="input1", path="path1", used_at=moment, size=0)
        f1.type = self.input_entry

        f2 = IOFilePut(name="input1", path="path1", used_at=moment, size=0)
        f2.type = self.input_entry

        toolwrapper1 = FooWrapper2(rule_name="rule1")
        toolwrapper1.files.append(f1)
        toolwrapper1.tables.append(t1)

        toolwrapper2 = FooWrapper2(rule_name="rule1")
        toolwrapper2.files.append(f2)
        toolwrapper2.tables.append(t2)

        t3 = IODbPut(model="FooBase", tablename="FooBase")
        t3.set_table(FooBase)
        t3.type = self.input_entry
        modif = ModificationTable(table_name="FooBase", date=moment)
        modif.tables.append(t3)

        f3 = IOFilePut(name="input1", path="path1", used_at=datetime.datetime.fromtimestamp(time.time()), size=0)
        f3.type = self.input_entry

        toolwrapper3 = FooWrapper2(rule_name="rule1")
        toolwrapper3.files.append(f3)
        toolwrapper3.tables.append(t3)

        self.assertTrue(toolwrapper1.same_input_than(toolwrapper2))
        self.assertFalse(toolwrapper1.same_input_than(toolwrapper3))

    def test_is_output_ok(self):

        moment = datetime.datetime.fromtimestamp(time.time())
        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.type = self.input_entry
        t1.used_at = moment
        modif = ModificationTable(table_name="FooBase", date=moment)
        modif.tables.append(t1)

        root = PathFinder.get_module_path()
        path_f1 = os.path.join(root, "test/output/path1")
        time.sleep(2)
        p = subprocess.Popen(["touch", path_f1])
        p.wait()

        f1 = IOFilePut(name="input1", path=path_f1, used_at=datetime.datetime.fromtimestamp(os.path.getmtime(path_f1)),
                       size=os.path.getsize(path_f1))
        f1.type = self.output_entry

        toolwrapper1 = FooWrapper2(rule_name="rule1")
        toolwrapper1.files.append(f1)
        toolwrapper1.tables.append(t1)


        f1 = IOFilePut(name="input1", path=path_f1, used_at=datetime.datetime.fromtimestamp(os.path.getmtime(path_f1)),
                       size=os.path.getsize(path_f1))

        f1.type = self.output_entry

        moment = datetime.datetime.fromtimestamp(time.time())
        t1 = IODbPut(model="FooBase", tablename="FooBase")
        t1.set_table(FooBase)
        t1.type = self.input_entry
        t1.used_at = moment
        modif = ModificationTable(table_name="FooBase", date=moment)
        modif.tables.append(t1)

        toolwrapper2 = FooWrapper2(rule_name="rule1")
        toolwrapper2.files.append(f1)
        toolwrapper2.tables.append(t1)

        # self.assertTrue(toolwrapper1.is_output_ok()) # TODO AG: Ask LG. Where is the is_output_ok function?
        # self.assertFalse(toolwrapper2.is_output_ok()) # TODO AG: Ask LG. Where is the is_output_ok function?

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove(os.path.join(self.s_root_path, "test/output"))
        OptionManager._drop()
        SQLManager._drop()

if __name__ == '__main__':
    unittest.main()



import os
import shutil
import errno

from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class ExampleBuilder:
    """
    This class is used to build the examples for the tutorials.

    Each method is supposed to build one distinct example.
    """

    def build(self):
        self.build_file_architecture()

    def build_file_architecture(self):
        """
        This builds the car example.
        """
        cwd = os.path.join(OptionManager.instance()["--directory"], "wopmars_example")
        example_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), "wopmars_example")

        # copy the folder wopmars_example in cwd
        ExampleBuilder.copy(example_directory, cwd)

        # empty.txt is an empty text file used in order to take into account the output directory
        os.remove(os.path.join(cwd, "output/empty.txt"))

    def build_snp(self):
        self.build_file_architecture_snp()

    def build_file_architecture_snp(self):
        """
        This builds the snp example.
        """
        cwd = os.path.join(OptionManager.instance()["--directory"], "wopmars_example_snp")
        example_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), "wopmars_example_snp")

        # copy the folder wopmars_example in cwd
        ExampleBuilder.copy(example_directory, cwd)

        # empty.txt is an empty text file used in order to take into account the output directory
        os.remove(os.path.join(cwd, "output/empty.txt"))

    @staticmethod
    # http://pythoncentral.io/how-to-recursively-copy-a-directory-folder-in-python/
    def copy(src, dest):
        try:
            shutil.copytree(src, dest, ignore=shutil.ignore_patterns('__pycache__'))
        except OSError as e:
            # If the error was caused because the source wasn't a directory
            if e.errno == errno.ENOTDIR:
                shutil.copy(src, dest)
            else:
                raise WopMarsException("Error while building the example", 'Directory not copied. Error: %s' % e)



from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from wopmars.framework.database.Base import Base


class Type(Base):
    """
    This class is the model of the table ``wom_type``. It stores the two kind of entry named "input" and "output" which
    are associated with an ID.

    Fields:

    - id: INTEGER - primary key -arbitrary ID
    - name: VARCHAR(255) - the name of the type ("input" or "output")
    """
    __tablename__ = "wom_type"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))

    # One type is in Many table
    tables = relationship("IODbPut", back_populates="type")
    # One type is in Many files
    files = relationship("IOFilePut", back_populates="type")

    def __repr__(self):
        return "<Type: %s>" % self.name



from termcolor import colored


class ColorPrint:
    """
    Static class which allow to return colored text in order to be printed
    """

    @staticmethod
    def yellow(text):
        return colored(text, 'yellow', attrs=['bold'])

    @staticmethod
    def green(text):
        return colored(text, 'green', attrs=['bold'])

    @staticmethod
    def red(text):
        return colored(text, 'red', attrs=['bold'])

    @staticmethod
    def cyan(text):
        return colored(text, 'cyan', attrs=['bold'])

    @staticmethod
    def blue(text):
        return colored(text, 'blue', attrs=['bold'])

    @staticmethod
    def magenta(text):
        return colored(text, 'magenta', attrs=['bold'])



import os
import unittest
from unittest import TestCase

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.management.WorkflowManager import WorkflowManager
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder
from wopmars.utils.exceptions.WopMarsException import WopMarsException


class TestWorkflowManager(TestCase):

    def setUp(self):
        OptionManager.initial_test_setup()
        SQLManager.instance().create_all()
        s_root_path = PathFinder.get_module_path()

        self.__s_path_to_example_definition_file_finishing = os.path.join(s_root_path, "test/resource/wopfile/example_def_file1.yml")
        self.__s_path_to_example_definition_file_that_end_with_error = os.path.join(s_root_path, "test/resource/wopfile/example_def_file5_never_ready.yml")

        self.__wm = WorkflowManager()


    def test_erase_output(self):
        #todo LG
        pass

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove("test/output")
        OptionManager._drop()
        SQLManager._drop()

    def test_run(self):
        # OptionManager.instance()["--dot"] = None
        #
        # OptionManager.instance()["--wopfile"] = self.__s_path_to_example_definition_file_finishing
        # with self.assertRaises(SystemExit):
        #     self.__wm.run()

        OptionManager.instance()["--wopfile"] = self.__s_path_to_example_definition_file_that_end_with_error
        with self.assertRaises(WopMarsException):
            self.__wm.run()

if __name__ == '__main__':
    unittest.main()



"""
This module contains the SetUtils class.
"""


class SetUtils:
    """
    Class containing methods that are usefull for dealing with sets
    """
    @staticmethod
    def all_elm_of_one_set_in_one_other(one, other):
        """
        Check if all elements of one set is in an other. The build in functions cannot be used (set.issubset()) because
        they rely on the hash of elements and not on their values.

        :param one: set
        :param other: set
        :return: boolean: True if all elements of one are in the other
        """
        for elm in one:
            is_in = False
            for elm2 in other:
                if elm == elm2:
                    is_in = True
                    break # Useless to keep looking in the other set since we just want to see if elm occurs once
            if not is_in:
                return False
        return True




"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey

from wopmars.framework.database.Base import Base


class FooBase6(Base):
    """
    Documentation for the class
    """
    __tablename__ = "FooBase6"

    id = Column(Integer, primary_key=True)

    foobase3 = relationship("FooBase3", back_populates="foobase6")


from wopmars.framework.database.Base import Base

from sqlalchemy import Column, String, Integer, ForeignKey
from sqlalchemy import UniqueConstraint

class SNP2Phenotype(Base):
    __tablename__ = 'SNP2Phenotype'
    __table_args__ = (
            UniqueConstraint('snp_id', 'phenotype_id'),
            )

    id = Column(Integer, primary_key=True, autoincrement=True)
    snp_id = Column(Integer, ForeignKey("SNP.id"), nullable=False)
    phenotype_id = Column(Integer, ForeignKey("Phenotype.id"), nullable=False)




import os
import unittest
from unittest import TestCase

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IOFilePut import IOFilePut
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.PathFinder import PathFinder


class TestIOFilePut(TestCase):
    def setUp(self):
        OptionManager.initial_test_setup()

        self.s_root_path = PathFinder.get_module_path()
        s_path_to_example_existing_file = os.path.join(self.s_root_path, "test/resource/input_files/example_existing_file.txt")
        s_path_to_example_existing_file2 = os.path.join(self.s_root_path, "test/resource/input_files/example_existing_file2.txt")
        s_path_to_example_not_existing_file = os.path.join(self.s_root_path, "test/resource/input_files/example_not_existing_file.txt")

        self.__io_file_existing = IOFilePut(name="existing_file", path=s_path_to_example_existing_file)
        self.__io_file_existing2 = IOFilePut(name="existing_file", path=s_path_to_example_existing_file)
        self.__io_file_existing3 = IOFilePut(name="existing_file2", path=s_path_to_example_existing_file)
        self.__io_file_existing4 = IOFilePut(name="existing_file", path=s_path_to_example_existing_file2)
        self.__io_file_not_existing = IOFilePut(name="not_existing_file", path=s_path_to_example_not_existing_file)

    def test_is_ready(self):
        self.assertTrue(self.__io_file_existing.is_ready())
        self.assertFalse(self.__io_file_not_existing.is_ready())

    def test_eq(self):
        self.assertEqual(self.__io_file_existing, self.__io_file_existing2)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_existing3)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_existing4)
        self.assertNotEqual(self.__io_file_existing, self.__io_file_not_existing)

    def tearDown(self):
        SQLManager.instance().get_session().close()
        SQLManager.instance().drop_all()
        PathFinder.dir_content_remove(os.path.join(self.s_root_path, "test/output"))
        OptionManager._drop()
        SQLManager._drop()

if __name__ == "__main__":
    unittest.main()



import importlib
import datetime
import time

from sqlalchemy.exc import OperationalError
from sqlalchemy.sql.ddl import DDL

from wopmars.framework.database.Base import Base
from sqlalchemy import Column, Integer, String, ForeignKey, DateTime
from sqlalchemy.orm import relationship, reconstructor

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.IOPut import IOPut
from wopmars.framework.database.tables.ModificationTable import ModificationTable
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper
from wopmars.utils.Logger import Logger
from wopmars.framework.database.tables.Type import Type
from sqlalchemy.sql.functions import func

class IODbPut(IOPut, Base):
    """
    This class extends IOPut and is specific to the input or output tables. It is the model which store the references
    to the actual tables needed by the user. The table ``wom_table`` associated with this model contains the
    following fields:

    - id: INTEGER - primary key - autoincrement - arbitrary ID
    - tablename: VARCHAR(255) - foreign key to the associated table: :class:`wopmars.framework.database.tables.ModificationTable.ModificationTable` - the name of the referenced table
    - model: VARCHAR(255) - the path to the model (in python notation)
    - rule_id: INTEGER - foreign key to the associated rule ID: :class:`wopmars.framework.database.tables.ToolWrapper.ToolWrapper`
    - type_id: INTEGER - foreign key to the associated type ID: :class:`wopmars.framework.database.tables.Type.Type`
    - used_at: DATE - date at which the table have been used
    """
    __tablename__ = "wom_table"

    id = Column(Integer, primary_key=True, autoincrement=True)
    tablename = Column(String(255), ForeignKey("wom_modification_table.table_name"))
    model = Column(String(255))
    rule_id = Column(Integer, ForeignKey("wom_rule.id"))
    type_id = Column(Integer, ForeignKey("wom_type.id"))
    used_at = Column(DateTime, nullable=True)

    # One table is in one rule
    rule = relationship("ToolWrapper", back_populates="tables", enable_typechecks=False)
    # One file has One type
    type = relationship("Type", back_populates="tables")

    modification = relationship("ModificationTable", back_populates="tables")

    # all the model names met since the begining of this instance of WopMaRS
    tablemodelnames = set()
    # al the table names met since the begining of this instance of WopMaRS
    tablenames = set()

    def __init__(self, model, tablename):
        """
        self.__table is initialized to None and will contain the model of this IODbPut object.

        :param model: The path to the model
        :type model: str
        :param tablename: The name of the table associated with the model
        :type tablename: str
        """
        # The file containing the table should be in PYTHONPATH
        Base.__init__(self, model=model, tablename=tablename)
        Logger.instance().debug(str(model) + " model loaded. Tablename: " + str(tablename))
        self.__table = None

    @reconstructor
    def init_on_load(self):
        """
        This is used by SQLAlchemy to regenerate the right object when loading it from the database. Here, we need to
        get back the actual Model from the model name and store it in self.__table.
        """
        for table in IODbPut.tablemodelnames:
            mod = importlib.import_module(table)
            try:
                if table == self.model:
                    # todo tabling
                    self.__table = eval("mod." + self.model.split(".")[-1])
            except AttributeError as e:
                raise e
        Logger.instance().debug(self.tablename + " table class reloaded. Model: " + self.model)

    def set_table(self, model):
        self.__table = model

    def get_table(self):
        return self.__table

    @staticmethod
    def create_triggers():
        """
        Create an INSERT, UPDATE, DELETE trigger on the tables created by the user in order to store the modifications date.
        """
        stmt = ["INSERT", "UPDATE", "DELETE"]
        for tablename in Base.metadata.tables:
            if tablename[:4] != "wom_":
                for s in stmt:
                    data={"statement": str(s), "tablename": str(tablename)} 
                    if SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'sqlite':
                        sql_trigger = """CREATE TRIGGER IF NOT EXISTS trigger_modif_%(tablename)s_%(statement)s AFTER %(statement)s ON %(tablename)s BEGIN UPDATE wom_modification_table SET date = CURRENT_TIMESTAMP WHERE table_name = '%(tablename)s'; END;"""%data
                    elif SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'mysql':
                        sql_trigger = """
CREATE TRIGGER IF NOT EXISTS modification_%(tablename)s_%(statement)s AFTER %(statement)s ON %(tablename)s for each row UPDATE wom_modification_table SET date = CURRENT_TIMESTAMP WHERE table_name = '%(tablename)s';
    """%data
                        obj_ddl = DDL(sql_trigger)
                        SQLManager.instance().create_trigger(Base.metadata.tables[tablename], obj_ddl)
                    elif SQLManager.instance().__dict__['d_database_config']['db_connection'] == 'postgresql':
                        sql_trigger = """
CREATE OR REPLACE FUNCTION modification_%(statement)s_%(tablename)s() RETURNS TRIGGER AS $modification_%(statement)s_%(tablename)s$
BEGIN
UPDATE wom_modification_table SET date = CURRENT_TIMESTAMP WHERE table_name = '%(tablename)s';
RETURN NULL; -- result is ignored since this is an AFTER trigger
END;
$modification_%(statement)s_%(tablename)s$ LANGUAGE plpgsql;
DROP TRIGGER IF EXISTS modification_%(statement)s_%(tablename)s ON "%(tablename)s";
CREATE TRIGGER modification_%(statement)s_%(tablename)s AFTER INSERT ON "%(tablename)s" FOR EACH ROW EXECUTE PROCEDURE modification_%(statement)s_%(tablename)s();
    """%data
                    # import pdb; pdb.set_trace()
                    obj_ddl = DDL(sql_trigger)
                    SQLManager.instance().create_trigger(Base.metadata.tables[tablename], obj_ddl)


    @staticmethod
    def set_tables_properties(tables):
        """
        Import the models of the current execution and then associate models with IODbPut objects.

        :param tables: the IODbPut which need their table properties to be set.
        :type tables: ResultSet(IODbPut)
        """
        # import models for avoid references errors between models when dealing with them
        IODbPut.import_models(set([t.model for t in tables]))

        for table in tables:
            # keep track of the models used in static variable of IODbPut
            IODbPut.tablemodelnames.add(table.model)
            # Associate model with the IODbPut object
            mod = importlib.import_module(table.model)
            table_model = eval("mod." + table.model.split(".")[-1])
            table.set_table(table_model)
            # keep track of table names used in static variable of IODbPut
            IODbPut.tablenames.add(table_model.__tablename__)
            SQLManager.instance().get_session().add(table)

    @staticmethod
    def get_execution_tables():
        """
        Return all the IODbPut objects found in model IODbPut.

        :return: ResultSet IODbPut objects
        """
        session = SQLManager.instance().get_session()
        execution_id = session.query(func.max(ToolWrapper.execution_id))
        return session.query(IODbPut).filter(IODbPut.rule_id == ToolWrapper.id).filter(ToolWrapper.execution_id == execution_id).all()

    @staticmethod
    def import_models(model_names):
        """
        Import all the given models

        :param model_names: The path to the models
        :type model_names: Iterable(String)
        """
        for t in model_names:
            Logger.instance().debug("IODbPut.import_models: importing " + str(t))
            importlib.import_module(t)

    def is_ready(self):
        """
        A IODbPut object is ready if its table exists and contains entries.

        :return: bool if the table is ready
        """
        session = SQLManager.instance().get_session()
        try:
            results = session.query(self.__table).first()
            if results is None:
                Logger.instance().debug("The table " + self.tablename + " is empty.")
                return False
        except OperationalError as e:
            Logger.instance().debug("The table " + self.__table.__tablename__ + " doesn't exist.")
            return False
        except Exception as e:
            session.rollback()
            raise e
            # todo twthread
        return True

    def __eq__(self, other):
        """
        Two IODbPut object are equals if their table attributes belongs to the same class and if the associated table
        has the same content

        :param other: IODbPut
        :return: boolean: True if the table attributes are the same, False if not
        """
        session = SQLManager.instance().get_session()
        if self.model != other.model or self.tablename != other.tablename:
            return False
        try:
            self_results = set(session.query(self.__table).all())
            other_results = set(session.query(other.get_table()).all())
            if self_results != other_results:
                return False
        except Exception as e:
            session.rollback()
            raise e
        return True

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return "<Table (" + self.type.name + "  ):\"" + str(self.tablename) + "\"; used_at:" + str(self.used_at) + ">"

    def __str__(self):
        return "<Table: " + self.tablename + "; model: " + self.model + ">"



"""
Module containing the Logger class.
"""
import logging
from logging.handlers import RotatingFileHandler

from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.Singleton import SingletonMixin
from wopmars.utils.ColorPrint import ColorPrint


class Logger(SingletonMixin):
    """
    The logger class allows to write in standard output and rotating files the log of executions

    There is 4 logging handler:
      - stream_handler: the logger which will be called to write in std out.
      - stream_handler_err: the logger which will be called to write errors in std out.
      - file_handler: the logger which will be called to write in the .log rotating file.
      - file_handler_err: the logger logger which will write everything in the .err rotating file.

    The file_handler_err should be used for debuging purposes.
    """

    def __init__(self):
        # the top level logger which will distribute messages to different handlers
        self.__logger = logging.getLogger("wopmars")
        self.__logger.setLevel(logging.DEBUG)
        formatter = logging.Formatter('%(asctime)s :: %(levelname)s :: %(message)s')

        # the loger for std out
        self.__stream_handler = logging.StreamHandler()

        # the loger which will write errors in stdout anyway
        self.__stream_handler_err = logging.StreamHandler()
        self.__stream_handler_err.setLevel(logging.WARNING)

        s_path_log_file = OptionManager.instance()["--log"].rsplit(".", 1)[0]
        # log file in append mode of size 1 Mo and 1 backup
        # handler equivalent to stream_handler in term of logging level but write in .log file
        self.__file_handler = RotatingFileHandler(s_path_log_file + ".log", 'a', 1000000, 1)
        formatter_file = logging.Formatter('%(asctime)s :: %(levelname)s :: %(message)s')
        self.__file_handler.setFormatter(formatter_file)

        # err file in append mode of size 1 Mo and 1 backup
        # this handler will write everything in the .err file.
        self.__err_handler = RotatingFileHandler(s_path_log_file + ".err", 'a', 1000000, 1)
        formatter_err = logging.Formatter('%(asctime)s :: %(message)s')
        self.__err_handler.setFormatter(formatter_err)
        self.__err_handler.setLevel(logging.DEBUG)

        verbosity = int(OptionManager.instance()["-v"])

        # set the verbosity of the stream handler and file handler depending of the needs of the user.
        if verbosity <= 0:
            self.__stream_handler.setLevel(logging.WARNING)
            self.__file_handler.setLevel(logging.WARNING)
        elif verbosity == 1:
            self.__stream_handler.setLevel(logging.INFO)
            self.__file_handler.setLevel(logging.INFO)
        elif verbosity >= 2:
            self.__stream_handler.setLevel(logging.DEBUG)
            self.__file_handler.setLevel(logging.DEBUG)

        # if printtools is demanded, things about execution will be printed out in std out
        if OptionManager.instance()["--printtools"]:
            self.__logger.addHandler(self.__stream_handler)
        else:
            self.__logger.addHandler(self.__stream_handler_err)
        self.__logger.addHandler(self.__file_handler)
        self.__logger.addHandler(self.__err_handler)

        self.__tw_logger = logging.getLogger("tw")
        self.__tw_streamhandler = logging.StreamHandler()
        self.__tw_logger.addHandler(self.__tw_streamhandler)
        self.__tw_logger.setLevel(logging.DEBUG)

    def info(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.blue('%(levelname)s :: %(asctime)s :: %(message)s'))
        self.__stream_handler.setFormatter(formatter_stream)
        self.__stream_handler_err.setFormatter(formatter_stream)

        self.__logger.info(msg)

    def debug(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.yellow('%(levelname)s :: %(asctime)s :: %(message)s'))
        self.__stream_handler.setFormatter(formatter_stream)
        self.__stream_handler_err.setFormatter(formatter_stream)

        self.__logger.debug(msg)

    def error(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.red('%(levelname)s :: %(asctime)s :: %(message)s'))
        self.__stream_handler.setFormatter(formatter_stream)
        self.__stream_handler_err.setFormatter(formatter_stream)

        self.__logger.error(msg)

    def warning(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.magenta('%(levelname)s :: %(asctime)s :: %(message)s'))
        self.__stream_handler.setFormatter(formatter_stream)
        self.__stream_handler_err.setFormatter(formatter_stream)

        self.__logger.warning(msg)

    def critical(self, msg):
        formatter_stream = logging.Formatter(ColorPrint.red('%(levelname)s :: %(asctime)s :: %(message)s'))
        self.__stream_handler.setFormatter(formatter_stream)

        self.__logger.critical(msg)

    def toolwrapper_debug(self, msg, tw_name):
        if OptionManager.instance()["--toolwrapper-log"]:
            self.__tw_streamhandler.setFormatter(
                logging.Formatter(ColorPrint.green(tw_name + ' ::  %(levelname)s :: %(asctime)s :: %(message)s')))
            self.__tw_logger.debug(msg)

    def toolwrapper_info(self, msg, tw_name):
        if OptionManager.instance()["--toolwrapper-log"]:
            self.__tw_streamhandler.setFormatter(
                logging.Formatter(ColorPrint.green(tw_name + ' :: %(levelname)s :: %(asctime)s :: %(message)s')))
            self.__tw_logger.info(msg)

    def toolwrapper_warning(self, msg, tw_name):
        if OptionManager.instance()["--toolwrapper-log"]:
            self.__tw_streamhandler.setFormatter(
                logging.Formatter(ColorPrint.green(tw_name + ' :: %(levelname)s :: %(asctime)s :: %(message)s')))
            self.__tw_logger.warning(msg)

    def toolwrapper_error(self, msg, tw_name):
        if OptionManager.instance()["--toolwrapper-log"]:
            self.__tw_streamhandler.setFormatter(logging.Formatter(ColorPrint.green(tw_name + ' :: %(asctime)s :: %(levelname)s :: %(message)s')))
            self.__tw_logger.error(msg)




"""
Module containing the Parser class
"""
import sys

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.functions import func

from wopmars.framework.database.SQLManager import SQLManager
from wopmars.framework.database.tables.Execution import Execution
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper
from wopmars.framework.management.DAG import DAG
from wopmars.framework.parsing.Reader import Reader
from wopmars.utils.Logger import Logger
from wopmars.utils.OptionManager import OptionManager
from wopmars.utils.exceptions.WopMarsException import WopMarsException

from networkx.algorithms.dag import is_directed_acyclic_graph


class Parser:
    """
    The Parser is used to organize the parsing of the Workflow Definition File.

    The aim of the Parser is to send the DAG representing the execution graph.
    """
    def __init__(self):
        """
        The constructor of Parser.

        Initialize the reader.
        
        :return:
        """
        self.__reader = Reader()

    def parse(self):
        """
        Organize the parsing of the Workflow Definition File or the Tool if only one tool is provided thanks to the
        tool command.

        Call the "read()" or the "load_one_toolwrapper" (depending on the use or not of tool command) method of the
        reader to insert in database the set of objects of the workflow.

        Then, the toolwrappers of the last execution are got back before calling the dag to build itself from the set of tools.

        The DAG is checked to actually being a Directed Acyclic Graph.

        If The "--dot" option is set, the dot and ps file are wrote here.

        :raise: WopMarsParsingException if the workflow is not a DAG.
        :return: the DAG
        """
        if not OptionManager.instance()["tool"]:
            self.__reader.read(OptionManager.instance()["--wopfile"])
        else:
            self.__reader.load_one_toolwrapper(OptionManager.instance()["TOOLWRAPPER"],
                                               OptionManager.instance()["--input"],
                                               OptionManager.instance()["--output"],
                                               OptionManager.instance()["--params"])
        # Get back the set of toolwrappers of the workflow before executing them.
        set_toolwrappers = self.get_set_toolwrappers()
        dag_tools = DAG(set_toolwrappers)
        if not is_directed_acyclic_graph(dag_tools):
            # todo find out the loop to specify it in the error message
            raise WopMarsException("Error while parsing the configuration file: \n\tThe workflow is malformed:",
                                   "The specified Workflow cannot be represented as a DAG.")
        s_dot_option = OptionManager.instance()["--dot"]
        if s_dot_option:
            Logger.instance().info("Writing the dot and ps files representing the workflow at " + str(s_dot_option))
            dag_tools.write_dot(s_dot_option)
            Logger.instance().debug("Dot and ps file wrote.")
        return dag_tools

    @staticmethod
    def get_set_toolwrappers():
        """
        Ask the database for toolwrappers of the current execution.

        The current execution is defined as the one with the highest id (it is auto_incrementing)

        :return: Set([ToolWrapper]) the set of toolwrappers of the current execution.
        """
        session = SQLManager.instance().get_session()
        set_toolwrappers = set()
        try:
            # query asking the db for the highest execution id
            execution_id = session.query(func.max(ToolWrapper.execution_id))
            Logger.instance().debug("Getting toolwrappers of the current execution. id = " + str(execution_id.one()[0]))
            set_toolwrappers = set(session.query(ToolWrapper).filter(ToolWrapper.execution_id == execution_id).all())
        except NoResultFound as e:
            raise e
        return set_toolwrappers



from sqlalchemy.sql.schema import ForeignKey

from wopmars.framework.database.Base import Base

from sqlalchemy import Column, Integer, String, Float


class PieceCar(Base):
    __tablename__ = "piece_car"

    id = Column(Integer, primary_key=True, autoincrement=True)
    car_serial_number = Column(String, unique=True)
    bodywork_serial_number = Column(String, ForeignKey("piece.serial_number"))
    engine_serial_number = Column(String, ForeignKey("piece.serial_number"))
    wheel_serial_number = Column(String, ForeignKey("piece.serial_number"))
    price = Column(Float)



# -*- coding: UTF-8 -*-

__author__ = "Luc Giffon"
__copyright__ = "Copyright 2017, Luc Giffon"
__email__ = "luc.giffon@gmail.com"
__license__ = "MIT"

import wopmars
from codecs import open
from os import path
import sys
import os
from configparser import RawConfigParser


def get_version():
    """Return package version from setup.cfg."""
    config = RawConfigParser()
    config.read(os.path.join('.', 'setup.cfg'))
    return config.get('metadata', 'version')

if sys.version_info < (3, 3):
    print("At least Python 3.3 is required.\n", file=sys.stderr)
    exit(1)

try:
    from setuptools import setup, find_packages
except ImportError:
    print("Please install setuptools before installing wopmars.",
          file=sys.stderr)
    exit(1)

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='wopmars',
    version=str(get_version()),
    description='Workflow Python Manager for Reproducible Science',
    long_description=long_description,
    url='https://github.com/aitgon/wopmars',
    author=__author__,
    author_email=__email__,
    license=__license__,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Database :: Front-Ends',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Software Development',
    ],
    download_url='https://github.com/aitgon/wopmars/archive/%s.tar.gz'%str(get_version()),
    keywords='workflow manager python object-oriented reproducible science database framework',
    packages=find_packages(exclude=['log', 'doc', '*.test_bak.*', 'soutenance', 'rapport']),
    install_requires = ["SQLAlchemy>=1.1.11", "docopt>=0.6.2", "schema>=0.6.5", "termcolor>=1.1.0", "PyYAML>=3.12", "networkx>=2.1"],
    extras_require={'pygraphviz': ['pygraphviz',]},
    data_files=[],
    entry_points={
        'console_scripts':['wopmars=wopmars:run']
    },
    package_data={},
    include_package_data=True
)




import unittest
from unittest import TestCase

from wopmars.utils.PathFinder import PathFinder


class TestPathFinder(TestCase):
    #def test_find_src(self):
    #    self.assertEqual(PathFinder.get_module_path("/home/user/wopmars/toto/tata"), "/home/user/wopmars/")

    def test_check_valid_path(self):
        self.assertIsNone(PathFinder.check_valid_path(None))
        self.assertRaises(FileNotFoundError, PathFinder.check_valid_path, ("/coucou/toto"))
        self.assertEqual(PathFinder.check_valid_path("/tmp/test_bak"), "/tmp/test_bak")


if __name__ == '__main__':
    unittest.main()



from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

# import models
import wopexamplesnp.model.SNP2Phenotype

import pandas
from sqlalchemy import select,insert
from sqlalchemy.sql.expression import bindparam
import csv

class InsertSNP2Phenotype(ToolWrapper): # inherit WopMars
    __mapper_args__ = {'polymorphic_identity': __name__}
    #
    __input_file_snp2phenotype = "snp2phenotype"
    __input_table_snp = "SNP"
    __input_table_phenotype = "Phenotype"
    __output_table_snp2phenotype = "SNP2Phenotype"

    def specify_input_file(self):
        return [InsertSNP2Phenotype.__input_file_snp2phenotype]

    def specify_input_table(self):
        return [InsertSNP2Phenotype.__input_table_snp,
                InsertSNP2Phenotype.__input_table_phenotype]

    def specify_output_table(self):
        return [InsertSNP2Phenotype.__output_table_snp2phenotype]

    def run(self):
        session = self.session()
        engine = session._WopMarsSession__session.bind
        conn = engine.connect()
        #
        snp2phenotype_path = self.input_file(InsertSNP2Phenotype.__input_file_snp2phenotype)
        snp_model = self.input_table(InsertSNP2Phenotype.__input_table_snp)
        phenotype_model = self.input_table(InsertSNP2Phenotype.__input_table_phenotype)
        snp2phenotype_model = self.output_table(InsertSNP2Phenotype.__output_table_snp2phenotype)
        snp2phenotype_df = pandas.read_table(snp2phenotype_path, header=None)
        #
        # read input file
        input_file_obj_list = []
        for line in csv.reader(open(snp2phenotype_path, 'r', encoding='utf-8'), delimiter="\t"):
            snp_rsid = int(line[0])
            phenotype_name = line[1]
            input_file_obj_list.append({'snp_rsid' : snp_rsid, 'phenotype_name' : phenotype_name})
        #
        # create insert
        snp_select = select([snp_model.id]).where(snp_model.rsid==bindparam('snp_rsid'))
        phenotype_select = select([phenotype_model.id]).where(phenotype_model.name==bindparam('phenotype_name'))
        output_table_insert = insert(table=snp2phenotype_model.__table__, values={'snp_id': snp_select, 'phenotype_id': phenotype_select})
        #
        if len(input_file_obj_list) > 0:
            if str(engine.__dict__['url']).split("://")[0]=='sqlite':
                engine.execute(output_table_insert.prefix_with("OR IGNORE", dialect='sqlite'), input_file_obj_list)
            elif str(engine.__dict__['url']).split("://")[0]=='mysql':
                    from warnings import filterwarnings # three lines to suppress mysql warnings
                    import MySQLdb as Database
                    filterwarnings('ignore', category = Database.Warning)
                    engine.execute(output_table_insert.prefix_with("IGNORE", dialect='mysql'), input_file_obj_list)
            elif str(engine.__dict__['url']).split("://")[0]=='postgresql':
                from sqlalchemy.dialects.postgresql import insert as pg_insert
                output_table_insert_pg = pg_insert(table=snp2phenotype_model.__table__, values={'snp_id': snp_select, 'phenotype_id': phenotype_select}).on_conflict_do_nothing(index_elements=['snp_id', 'phenotype_id'])
                engine.execute(output_table_insert_pg, input_file_obj_list)




from wopmars.framework.database.tables.ToolWrapper import ToolWrapper

# import models
import wopexamplesnp.model.SNP

import csv

class InsertPhenotype(ToolWrapper): # inherit WopMars
    __mapper_args__ = {'polymorphic_identity': __name__}
    __input_file_snp2phenotype = "snp2phenotype"
    __output_table_phenotype = "Phenotype"

    def specify_input_file(self):
        return [InsertPhenotype.__input_file_snp2phenotype]

    def specify_output_table(self):
        return [InsertPhenotype.__output_table_phenotype]

    def run(self):
        session = self.session()
        engine = session._WopMarsSession__session.bind
        conn = engine.connect()
        #
        snp2phenotype_path = self.input_file(InsertPhenotype.__input_file_snp2phenotype)
        phenotype_model = self.output_table(InsertPhenotype.__output_table_phenotype)
        #
        # read input file
        input_file_obj_list = []
        for line in csv.reader(open(snp2phenotype_path, 'r', encoding='utf-8'), delimiter="\t"):
            phenotype_name = line[1]
            input_file_obj = {'name': phenotype_name}
            input_file_obj_list.append(input_file_obj)
        #
        # insert input_file_obj_list
        if len(input_file_obj_list) > 0:
            if str(engine.__dict__['url']).split("://")[0]=='sqlite':
                engine.execute(phenotype_model.__table__.insert().prefix_with("OR IGNORE"), input_file_obj_list)
            elif str(engine.__dict__['url']).split("://")[0]=='mysql':
                    from warnings import filterwarnings # three lines to suppress mysql warnings
                    import MySQLdb as Database
                    filterwarnings('ignore', category = Database.Warning)
                    engine.execute(phenotype_model.__table__.insert().prefix_with("IGNORE"), input_file_obj_list)
            elif str(engine.__dict__['url']).split("://")[0]=='postgresql':
                from sqlalchemy.dialects.postgresql import insert as pg_insert
                engine.execute(pg_insert(phenotype_model.__table__).on_conflict_do_nothing(index_elements=['name']), input_file_obj_list)
            else:
                raise "Error: This engine is not implemented."




"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer, String

from wopmars.framework.database.Base import Base


class FooBasePackaged(Base):
    """
    Documentation for the class
    """
    __tablename__ = "FooBasePackaged"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String(255))

    def __repr__(self):
        s = ""
        s += "<FooBasePackaged (id: %s; name: %s)>" % (self.id, self.name)
        return s



from wopmars.framework.database.Base import Base

from sqlalchemy import Column, String, Integer, SmallInteger
from sqlalchemy import UniqueConstraint

class SNP(Base):
    __tablename__ = 'SNP'
    __table_args__ = (
            UniqueConstraint('chrom', 'position'),
            UniqueConstraint('rsid'),
            )

    id = Column(Integer, primary_key=True, autoincrement=True)
    chrom = Column(SmallInteger, nullable=False)
    position = Column(Integer, nullable=False)
    rsid = Column(Integer, nullable=False)




import threading

from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class Query(ToolWrapper):
    __mapper_args__ = {'polymorphic_identity': "sprintFive.Query"}

    def specify_input_table(self):
        return ["FooBase"]

    def run(self):
        print(threading._active)
        self.session().query(self.input_table("FooBase")).all()



"""
Example of module documentation which can be
multiple-lined
"""
from sqlalchemy import Column, Integer
from sqlalchemy.orm import relationship
from sqlalchemy.sql.schema import ForeignKey

from wopmars.framework.database.Base import Base


class FooBase4(Base):
    """
    Documentation for the class
    """
    __tablename__ = "FooBase4"

    id = Column(Integer, primary_key=True)
    id_foobase3 = Column(Integer, ForeignKey("FooBase3.id"))
    id_foobase5 = Column(Integer, ForeignKey("FooBase5.id"))

    foobase3 = relationship("FooBase3", back_populates="foobase4")
    foobase5 = relationship("FooBase5", back_populates="foobase4")


"""
Module containing the FooWrapper2 class
"""
from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class FooWrapper3(ToolWrapper):
    """
    This class has been done for example/testing purpose.
    Modifications may lead to failure in tests.
    """
    def specify_input_file(self):
        return ["input1"]

    def specify_input_table(self):
        return ["FooBase"]

    def specify_output_file(self):
        return ["output1"]

    def specify_output_table(self):
        return ["FooBase"]

    def specify_params(self):
        return {"param1": "float|required"}



import random
import string

from wopmars.framework.database.tables.ToolWrapper import ToolWrapper


class CarAssembler(ToolWrapper):
    __mapper_args__ = {
        "polymorphic_identity": "wopmars.example.CarAssembler"
    }

    def specify_output_file(self):
        if not self.option("to_file"):
            return []
        else:
            return ["piece_car"]

    def specify_input_table(self):
        return ["piece"]

    def specify_output_table(self):
        if self.option("to_file"):
            return []
        else:
            return ["piece_car"]

    def specify_params(self):
        return {
            "to_file": "bool",
            "max_price": "int",
        }

    def run(self):
        session = self.session()
        Piece = self.input_table("piece")
        if not self.option("to_file"):
            Piece_car = self.output_table("piece_car")

        max_price = self.option("max_price")

        wheels = session.query(Piece).filter(Piece.type == "wheel").all()
        engines = session.query(Piece).filter(Piece.type == "engine").all()
        bodyworks = session.query(Piece).filter(Piece.type == "bodywork").all()
        uniques_serials = [s[0] for s in session.query(Piece.serial_number).all()]

        s = "car_serial_number, bodywork_serial_number, engine_serial_number, wheel_serial_number, price\n"

        for w in wheels:
            for e in engines:
                for b in bodyworks:
                    car_serial_number = CarAssembler.id_generator()
                    while car_serial_number in uniques_serials:
                        car_serial_number = CarAssembler.id_generator()
                    uniques_serials.append(car_serial_number)
                    price = w.price + e.price + b.price
                    if not max_price or (max_price and price <= max_price):
                        if not self.option("to_file"):
                            session.add(Piece_car(car_serial_number=car_serial_number,
                                                  bodywork_serial_number=b.serial_number,
                                                  engine_serial_number=e.serial_number,
                                                  wheel_serial_number=w.serial_number,
                                                  price=price))
                        s += ";".join([car_serial_number,
                                       b.serial_number,
                                       e.serial_number,
                                       w.serial_number,
                                       str(price)]) + "\n"

        if self.option("to_file"):
            file_to_write = open(self.output_file("piece_car"), 'w')
            file_to_write.write(s)
            file_to_write.close()
        else:
            session.commit()

    @staticmethod
    def id_generator(size=10, chars=string.ascii_uppercase + string.digits):
        return ''.join(random.choice(chars) for _ in range(size))
